import builtins
import sys


def noop(*args):
    """A no-op function.

    Args:
        *args: The arguments to ignore
    """
    pass


original_import = builtins.__import__


def custom_import(name, globals=None, locals=None, fromlist=(), level=0) -> object:
    """A custom import to disable reflex_ai.enable.

    Args:
        name: The name of the module to import.
        globals: The global namespace.
        locals: The local namespace.
        fromlist: The list of names to import.
        level: The level in the package hierarchy.

    Returns:
        The imported module.
    """
    module = original_import(name, globals, locals, fromlist, level)

    # Disable reflex_ai.enable
    if name == "reflex_ai" and fromlist and "enable" in fromlist:
        module.enable = noop
    return module


def validate_source(source_file: str, validation_function: str):
    """Validate the source code by running the validation function.

    Args:
        source_file: The source code to validate.
        validation_function: The name of the validation function.
    """
    builtins.__import__ = custom_import
    try:
        # Read the source code
        with open(source_file) as f:
            source_to_exec = f.read()

        # Execute the source code.
        env = {}
        exec(source_to_exec, env, env)

        # Run the validation function.
        component_fn = env[validation_function]
        eval(f"{component_fn.__name__}()", env, env)
    finally:
        builtins.__import__ = original_import


if __name__ == "__main__":
    validate_source(sys.argv[1], sys.argv[2])



import reflex as rx
from reflex_ai.toolbar import ToolbarState
from reflex_ai.views.accordion.accordion import accordion
from reflex_ai.components.styles import get_main_styles


class SettingsState(rx.State):
    is_open: bool = False
    side: str = "right"
    color: str = "violet"
    opacity: list[float] = [1.0]

    def toggle_side(self):
        self.side = "left" if self.side == "right" else "right"

    def set_opacity(self, value: list):
        self.opacity = value


def hide_button_outter() -> rx.Component:
    return rx.el.button(
        rx.icon(
            tag="panel-right-open",
            stroke_width="1.5",
            class_name="text-[--slate-11] shrink-0",
            rotate=rx.cond(SettingsState.side == "right", "0deg", "180deg"),
            size=20,
        ),
        on_click=SettingsState.setvar("is_open", ~SettingsState.is_open),
        class_name="top-1/2 inline-flex fixed justify-center items-center border-[--slate-5] bg-[--slate-1] hover:bg-[--slate-3] shadow-medium px-3 py-[0.275rem] border border-r-0 w-7 h-7 transition-bg cursor-pointer",
        border_radius=rx.cond(
            SettingsState.side == "left", "0 0.35rem 0.35rem 0", "0.35rem 0 0 0.35rem"
        ),
        right=rx.cond(SettingsState.side == "right", "0px", "auto"),
        left=rx.cond(SettingsState.side == "left", "0px", "auto"),
        display=rx.cond(SettingsState.is_open, "none", "flex"),
    )


def playground(page) -> rx.Component:
    return rx.box(
        page(),
        get_main_styles(),
        rx.el.style(
            f"""
:root {{
    --select-bg: var(--{SettingsState.color}-a4);
    --select-outline: var(--{SettingsState.color}-9);
}}
"""
        ),
        hide_button_outter(),
        rx.box(
            accordion(),
            transform=rx.cond(
                SettingsState.is_open,
                "translateX(0%)",
                rx.cond(
                    SettingsState.side == "right",
                    "translateX(100%)",
                    "translateX(-100%)",
                ),
            ),
            transition="transform 0.275s ease-out",
            opacity=SettingsState.opacity[0],
            class_name="z-[9999] fixed h-full overflow-hidden top-4 will-change-transform flex",
            right=rx.cond(
                SettingsState.side == "right",
                rx.cond(SettingsState.is_open, "1rem", "0px"),
                "auto",
            ),
            left=rx.cond(
                SettingsState.side == "left",
                rx.cond(SettingsState.is_open, "1rem", "0px"),
                "auto",
            ),
            justify_content=rx.cond(
                SettingsState.side == "right", "flex-end", "flex-start"
            ),
            pointer_events=rx.cond(SettingsState.is_open, "auto", "none"),
        ),
        on_mount=ToolbarState.load_diff,
    )



import ast
import importlib
import inspect
import subprocess
import sys

from reflex_ai import paths
from reflex_ai.scripts import get_module_script


def import_module_from_scratch(module_name: str, root_dir: str, scratch_dir: str):
    """Import the module from the scratchpad.

    Args:
        module_name: The name of the module to import.
        root_dir: The root directory of the module.
        scratch_dir: The scratch directory of the module.

    Returns:
        The imported module.
    """
    module = importlib.import_module(module_name)
    module_file = inspect.getsourcefile(module)
    module_tmp_file = module_file.replace(root_dir, scratch_dir)
    spec = importlib.util.spec_from_file_location(
        module.__name__,
        module_tmp_file,
    )
    new_module = importlib.util.module_from_spec(spec)
    # spec.loader.exec_module(new_module)
    return new_module


def get_module_source(module_name: str) -> str:
    """Get the file path of a module in a subprocess.

    Args:
        module_name: The name of the module.

    Returns:
        The source code of the module.
    """
    path = inspect.getfile(get_module_script)
    result = subprocess.run(
        [
            sys.executable,
            path,
            module_name,
            str(paths.base_paths[0].parent),
            str(paths.tmp_root_path),
        ],
        capture_output=True,
        text=True,
        check=True,
    )
    # Capture the output
    module_file = result.stdout.strip().split("\n")[-1]
    with open(module_file) as f:
        return f.read()


def get_module_ast(module_name: str) -> ast.Module:
    """Get the AST of a module.

    Args:
        module_name: The name of the module.

    Returns:
        The AST of the module.
    """
    # Import the module
    module_source = get_module_source(module_name)

    # Parse the module source code into an AST
    module_ast = ast.parse(module_source)
    return module_ast


def find_node(tree, element_type: str, element_name: str) -> ast.AST | None:
    for node in ast.walk(tree):
        if (
            isinstance(node, ast.FunctionDef)
            and element_type == "function"
            and node.name == element_name
        ):
            return node
        elif (
            isinstance(node, ast.ClassDef)
            and element_type == "class"
            and node.name == element_name
        ):
            return node
        elif isinstance(node, ast.Assign) and element_type == "variable":
            for target in node.targets:
                if isinstance(target, ast.Name) and target.id == element_name:
                    return node
        elif isinstance(node, ast.Return) and element_type == "return":
            return node
    return None



import reflex as rx
from reflex_ai.components.code_block import code_block
from reflex_ai.selection import ClickSelectionState


def selected_code_box() -> rx.Component:
    return rx.box(
        code_block(
            rx.cond(
                ClickSelectionState.code,
                ClickSelectionState.code,
                "Select a component to edit",
            )
        ),
        class_name="flex flex-col gap-3 w-full",
    )



"""Utility functions for the reflex_ai package."""

import os
import shutil
import subprocess
import sys
import tempfile
from pathlib import Path

import black
import difflib
import filecmp
import inspect

import reflex as rx
from reflex.utils import console

from reflex_ai import paths
from reflex_ai.scripts import validate_source as validate_source_script

SCRATCH_DIR_NAME = "reflex_ai_tmp"


def get_scratch_dir(app_dir: Path) -> Path:
    """Get the location of the scratch directory where the agent makes changes.

    Args:
        app_dir: The directory of the app that was copied into the scratch directory.

    Returns:
        The path to the scratch directory.
    """
    return app_dir.parent / rx.constants.Dirs.WEB / SCRATCH_DIR_NAME


def create_scratch_dir(app_dir: Path, overwrite: bool = False) -> Path:
    """Create a scratch directory for the agent to make changes to.

    Args:
        app_dir: The directory of the app to copy into the scratch directory.
        overwrite: Whether to overwrite the scratch directory if it already exists.

    Returns:
        The path to the created directory.
    """
    scratch_dir = get_scratch_dir(app_dir)

    # If the scratch directory already exists, skip.
    if scratch_dir.exists() and not overwrite:
        console.debug(
            f"Scratch directory already exists at {scratch_dir}. Skipping creation."
        )
        return scratch_dir

    # Copy the app directory to a temporary path for modifications.
    shutil.copytree(
        app_dir,
        scratch_dir / app_dir.name,
        dirs_exist_ok=True,
    )

    # Rename the copied directory to the scratch directory name.
    modify_scratch_directory(scratch_dir)
    return scratch_dir


def modify_scratch_directory(directory: Path) -> None:
    """Modify Python files in the scratch directory to use the correct state class.

    Args:
        directory: The directory containing files to modify.
    """
    for root, _, files in os.walk(directory):
        for file in files:
            if not file.endswith(".py") or "rxconfig" in file:
                continue
            file_path = Path(root) / file
            modify_scratch_file(file_path)


def modify_scratch_file(file_path: Path):
    """Modify the content of a single file in the scratch directory.

    Args:
        file_path: The path to the file to modify.
    """
    with open(file_path, "r") as f:
        content = f.read()

    content = content.replace("(rx.State)", "(EditableState)")
    content = f"from reflex_ai import EditableState\n{content}"

    console.debug(f"Writing to {file_path}")
    with open(file_path, "w") as f:
        f.write(content)


def commit_scratch_dir(app_dir: Path, files: list[str]) -> None:
    """Copy all files from the scratch directory back to the corresponding app directory.

    Args:
        app_dir: The original app directory to copy files back to.
        files: The list of files to copy back.
    """
    scratch_dir = get_scratch_dir(app_dir)

    # Iterate over the files and copy them back to the app directory.
    for file in files:
        relative_path = Path(file).relative_to(app_dir.parent)
        source_file = scratch_dir / relative_path
        target_file = Path(file)

        with open(source_file, "r") as f:
            content = f.read()

        content = content.replace("(EditableState)", "(rx.State)")
        content = content.replace("from reflex_ai import EditableState\n", "")

        console.debug(f"Writing to {target_file}")
        with open(target_file, "w") as f:
            f.write(content)


def format_code(code: str) -> str:
    """Format the code using black.

    Args:
        code: The code to format.

    Returns:
        The formatted code.
    """
    return black.format_str(code, mode=black.FileMode())


def diff_directories(dir1: str, dir2: str) -> dict[str, list[str]]:
    """Diff two directories and return the diffs for all Python files.

    Args:
        dir1: The first directory.
        dir2: The second directory.

    Returns:
        A dictionary mapping file paths to their diffs.
    """
    diffs: dict[str, list[str]] = {}

    def compare_dirs(dcmp: filecmp.dircmp) -> None:
        for name in dcmp.diff_files:
            if not name.endswith(".py"):
                continue
            file1 = Path(dcmp.left) / name
            file2 = Path(dcmp.right) / name
            with file1.open() as f1, file2.open() as f2:
                f1_lines = format_code(f1.read()).splitlines()
                f2_lines = format_code(
                    f2.read()
                    .replace("(EditableState)", "(rx.State)")
                    .replace("from reflex_ai import EditableState\n", "")
                ).splitlines()
                diff = list(
                    difflib.unified_diff(
                        f1_lines, f2_lines, fromfile=str(file1), tofile=str(file2)
                    )
                )
                diffs[str(file1)] = diff
        for sub_dcmp in dcmp.subdirs.values():
            compare_dirs(sub_dcmp)

    dirs_cmp = filecmp.dircmp(dir1, dir2)
    compare_dirs(dirs_cmp)
    return diffs


def directory_diff() -> dict[str, list[str]]:
    """Diff the scratchpad and the base directories.

    Returns:
        A dictionary mapping file paths to their diffs.

    Raises:
        AssertionError: If the paths are not properly configured.
    """
    assert (
        isinstance(paths.base_paths, list)
        and len(paths.base_paths) > 0
        and isinstance(paths.base_paths[0], Path)
    )
    assert isinstance(paths.tmp_root_path, Path)
    return diff_directories(str(paths.base_paths[0].parent), str(paths.tmp_root_path))


def validate_source(source: str, validation_function: str):
    """Validate the diff and write the changes to the file.

    Args:
        source: The source code to validate.
        validation_function: The function to validate the source code.
    """
    # Write the source to a temporary file.
    with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
        tmp_file.write(source.encode())
        path = inspect.getfile(validate_source_script)

    # Run the validation script in a subprocess.
    try:
        subprocess.run(
            [sys.executable, path, tmp_file.name, validation_function],
            capture_output=True,
            text=True,
            check=True,
        )
    finally:
        os.unlink(tmp_file.name)



"""The local Reflex agent."""

import importlib
import inspect
from pathlib import Path
import types
from types import FunctionType

import reflex as rx
import sys

from reflex_ai import paths
from reflex_ai.utils import path_utils

from reflex_ai.selection import ClickSelectionState


class EditableState(rx.State):
    """Rewrite rx.State imports in copied modules to this new class to avoid conflicts."""

    @classmethod
    def get_name(cls) -> str:
        """Get the name of the state.

        Returns:
            The name of the state.
        """
        module = cls.__module__.replace(".", "___")
        return rx.utils.format.to_snake_case(f"Editable___{module}___{cls.__name__}")


def enable(app: rx.App):
    """Enable the agent on an app.

    Args:
        app: The app to enable the agent on.

    Note:
        For now, this must be called before add_page is called as
        we override the add_page method.
    """
    from reflex.utils.exec import is_prod_mode
    from reflex import reflex

    # Ensure that user logs in before using the app.
    reflex._login()

    # Skip if in production mode.
    if is_prod_mode():
        return

    from .selection import clickable
    from .playground import playground

    # The base path is the directory where the app is defined.
    caller_frame = inspect.stack()[1]
    caller_path = caller_frame.filename

    # We care about the main directory, where the files live, and the parent directory, where .web lives
    base_paths = [Path(caller_path).parent, Path(caller_path).parent.parent]

    # Skip if the base path is a temporary directory.
    if path_utils.SCRATCH_DIR_NAME in str(base_paths[0]):
        return

    tmp_root_path = path_utils.create_scratch_dir(base_paths[0])
    paths.base_paths = base_paths
    paths.tmp_root_path = tmp_root_path
    filename = tmp_root_path / base_paths[0].name / "__init__.py"
    spec = importlib.util.spec_from_file_location(
        "tmp" + base_paths[0].name,
        filename,
    )
    new_module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(new_module)
    sys.modules["tmp" + base_paths[0].name] = new_module

    def add_page(self, component, *args, **kwargs):
        # Skip if the component is not a function.
        if not isinstance(component, FunctionType):
            return

        # Add the page normally.
        route = kwargs.pop("route", rx.utils.format.format_route(component.__name__))
        rx.App.add_page(self, component, *args, route=route, **kwargs)

        # Determine which module the component came from
        module = inspect.getmodule(component)
        if module is None:
            # Skip if the component does not come from a known module.
            return

        # Get the scratch module path.
        module_path = inspect.getfile(component)
        new_module_path = Path(
            module_path.replace(str(base_paths[0].parent), str(tmp_root_path))
        )
        spec = importlib.util.spec_from_file_location(
            "tmp" + module.__name__,
            new_module_path,
        )
        new_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(new_module)

        # Add the page on the edit route.
        rx.App.add_page(
            self,
            clickable(base_paths=base_paths)(
                lambda: playground(getattr(new_module, component.__name__))
            ),
            route=f"/{route}/edit" if route != "index" else "/edit",
            on_load=ClickSelectionState.click_page(
                new_module_path, component.__name__
            ).prevent_default.stop_propagation,
        )

    app.add_page = types.MethodType(add_page, app)



"""Handle point-and-click element selection in a Reflex app"""

import functools
import inspect
import textwrap
from pathlib import Path
from typing import Any
import reflex as rx

from reflex_ai import paths
from reflex_ai.utils import path_utils, ast_utils

# These components will never be selectable.
ignored_components = [
    rx.accordion.root.__self__,
    rx.accordion.icon.__self__,
    rx.accordion.item.__self__,
    rx.accordion.trigger.__self__,
    rx.alert_dialog.trigger.__self__,
    rx.hover_card.root.__self__,
    rx.hover_card.trigger.__self__,
    rx.hover_card.content.__self__,
    rx.popover.root.__self__,
    rx.popover.trigger.__self__,
    rx.popover.content.__self__,
    rx.recharts.responsive_container.__self__,
    rx.select.content.__self__,
    rx.select.group.__self__,
    rx.select.item.__self__,
    rx.select.root.__self__,
    rx.theme.__self__,
    rx.Fragment,
]
# Radix triggers that fire on_pointer_down events need to be disabled so that on_click works.
neuter_pointer_down = [
    rx.select.trigger.__self__,
]
# For components that do not support on_click at all, wrap these in a box.
box_wrap_components = [
    rx.markdown.__self__,
]
# Preserve a reference to the original create method's function.
og_component_create = rx.Component.create.__func__


class Selection(rx.Base):
    """Capture details of a selected component."""

    # The filename of the component.
    filename: str = ""

    # The function name of the component.
    function_name: str = ""

    # The starting line number of the component.
    start_line: int = 0

    # The ending line number of the component.
    end_line: int = 0

    # The list of modules for the call stack.
    modules: list[dict[str, Any]] = []

    def get_selection_id(self) -> str:
        """Get a unique identifier for the selection.

        Returns:
            A unique identifier for the selection.
        """
        return f"{self.filename}:{self.start_line}-{self.end_line}"

    def get_selection_code(self) -> str:
        """Get the code for the selection.

        Returns:
            The code for the selection.
        """
        return textwrap.dedent(
            "\n".join(
                Path(self.filename)
                .read_text()
                .splitlines()[self.start_line - 1 : self.end_line]
            )
        )

    def get_selection_module(self) -> str:
        """Get the module for the selection.

        Returns:
            The module for the selection.
        """
        if paths.tmp_root_path is None:
            return ""
        module_rel_path = (
            Path(self.filename).relative_to(paths.tmp_root_path).with_suffix("")
        )
        return ".".join(module_rel_path.parts)


class ClickSelectionState(rx.State):
    """Track the currently selected component and the code that created it."""

    # The currently selected component.
    selection: Selection = Selection()

    # The code that created the selected component.
    code: str = ""

    # The id of the currently selected element.
    selected_element: str = ""

    async def handle_click(self, selection: Selection):
        """Handle a click event on a component.

        Args:
            selection: The selected component.
        """
        # Importing here to avoid circular import
        from reflex_ai.playground import SettingsState

        # Open the toolbar
        settings_state = await self.get_state(SettingsState)
        settings_state.is_open = True

        # Set the selection.
        selection = Selection.parse_obj(selection)
        self.selection = selection
        self.selected_element = selection.get_selection_id()
        self.code = selection.get_selection_code()

    async def click_page(self, filename, function):
        """
        Given a file, and the "main component" of the file, handle a click event on the component.

        Args:
            filename: The file to open
            function: The "main component" of the file
        """

        import ast

        with open(filename, "r") as f:
            tree = ast.parse(f.read())

        main_page_node = ast_utils.find_node(tree, "function", function)
        return_statement = ast_utils.find_node(main_page_node, "return", None)

        selection = Selection(
            filename=filename,
            function_name=function,
            start_line=return_statement.lineno,
            end_line=getattr(return_statement, "end_lineno", return_statement.lineno),
            modules=[{"filename": filename, "function": function}],
        )

        await self.handle_click(selection)


def get_selection_props(selection: Selection) -> dict[str, Any]:
    """Generate on_click handler and outline props for selectable components.

    Args:
        selection: The selected component.

    Returns:
        A dictionary of props that includes the on_click handler and outline props.
    """
    return {
        "on_click": ClickSelectionState.handle_click(
            selection
        ).prevent_default.stop_propagation,
        "position": "relative",
        "::before": rx.cond(
            ClickSelectionState.selected_element == f"{selection.get_selection_id()}",
            {
                "content": "''",
                "position": "absolute",
                "top": "-4px",
                "left": "-4px",
                "width": "calc(100% + 8px)",
                "height": "calc(100% + 8px)",
                "backgroundColor": "var(--select-bg)",
                "outline": "2px dashed var(--select-outline)",
                "borderRadius": "0.35rem",
                "zIndex": "999",
                "pointerEvents": "none",
                "display": "block",
            },
            {},
        ),
    }


def component_create_override(base_paths: list[Path]) -> classmethod:
    """Generate an override for Component.create that is active for the given paths.

    Args:
        base_paths: Components originating from modules prefixed by these paths will be clickable.

    Returns:
        A replacement Component.create function that adds an on_click handler where needed.
    """
    base_paths = base_paths + [path_utils.SCRATCH_DIR_NAME]

    @classmethod
    def _component_create_override(cls, *children, **props):
        if cls not in ignored_components:
            # Walk up the stack to find the first frame that originates from a base path.
            stack = inspect.stack()
            for i, frame in enumerate(stack):
                # Skip the module frame.
                if frame.function == "<module>":
                    continue

                # If the frame originates from a base path, create the selection.
                if any(frame.filename.startswith(str(p)) for p in base_paths):
                    # The end index is the frame where the playground component is called.
                    end_idx = [frame.function for frame in stack].index("playground")

                    # Find all the modules in the call stack.
                    modules = [
                        {
                            "filename": frame.filename,
                            "function": frame.function,
                        }
                        for frame in stack[i:end_idx]
                    ]

                    # Create the selection.
                    selection = Selection(
                        filename=frame.filename,
                        function_name=frame.function,
                        start_line=frame.lineno,
                        end_line=frame.positions.end_lineno,
                        modules=modules,
                    )
                    down_props = get_selection_props(selection)

                    # If the component is a box wrap component, wrap it in a box.
                    if cls in box_wrap_components:
                        return rx.box(
                            og_component_create(cls, *children, **props),
                            **down_props,
                        )

                    # Add the selection props to the component.
                    props.update(down_props)

                    # If the component is a trigger that fires on_pointer_down, disable it.
                    if cls in neuter_pointer_down:
                        props.setdefault("special_props", set()).add(
                            rx.Var.create(
                                "onPointerDown={(e) => e.preventDefault()}",
                                _var_is_string=False,
                            ),
                        )
                    break
        return og_component_create(cls, *children, **props)

    return _component_create_override


def clickable(base_paths: list[Path] | None = None):
    """A decorator helper to make all components in a given page clickable to select.

    The active selection (filename and line range) is stored in
    ClickSelectionState.selected_element

    The code for the selection is cached in ClickSelectionState._selected_code

    Args:
        base_paths: Components originating from modules prefixed by these paths will be clickable.

    Returns:
        A decorator that adds an on_click handler to the component.
    """
    if base_paths is None:
        base_paths = [Path(".").resolve()]

    def outer(fn):
        @functools.wraps(fn)
        def inner(*args, **kwargs):
            # Override the component create method.
            rx.Component.create = component_create_override(base_paths=base_paths)
            try:
                # Call the page wrapper
                return fn(*args, **kwargs)
            finally:
                # Restore the original component create method.
                rx.Component.create = classmethod(og_component_create)

        # Preserve the original function name to avoid having to specify route in add_page.
        inner.__name__ = fn.__name__
        return inner

    return outer



import reflex as rx
from reflex_ai.views.accordion.selected_code_box import selected_code_box
from reflex_ai.views.accordion.process_box import process_box
from reflex_ai.views.accordion.diff_box import diff_box
from reflex_ai.toolbar import ToolbarState
from reflex_ai.views.bottom_bar import bottom_bar


def accordion_trigger(text: str, icon: str, value: str, index: int) -> rx.Component:
    return rx.accordion.trigger(
        rx.accordion.header(
            text,
            rx.box(
                rx.icon(
                    tag=icon,
                    size=17,
                    class_name="shrink-0",
                    color=rx.cond(
                        index >= ToolbarState.step, "var(--slate-11)", "white"
                    ),
                ),
                class_name="inline-flex absolute justify-center items-center rounded-full w-7 h-7 -start-[2.685rem]",
                background_color=rx.cond(
                    index >= ToolbarState.step, "var(--slate-4)", "var(--green-9)"
                ),
            ),
            class_name="relative font-base !leading-7 overflow-visible",
        ),
        rx.icon(
            tag="chevrons-up-down",
            size=20,
            stroke_width="1.5",
            class_name="group-data-[state=open]:text-[--slate-12] group-hover:text-[--slate-12] text-[--slate-11] transition-color",
        ),
        on_click=ToolbarState.setvar(
            "current_step", rx.cond(value == ToolbarState.current_step, "", value)
        ),
        class_name="justify-between items-center hover:!bg-[--slate-3] !px-3 !py-0 rounded-lg w-full !text-[--slate-12] transition-bg group",
    )


def accordion_item(
    text: str, icon: str, content: rx.Component, value: str, index: int
) -> rx.Component:
    return rx.accordion.item(
        accordion_trigger(text, icon, value, index),
        rx.accordion.content(
            content,
            class_name="relative !p-0 before:!h-4 after:!h-0",
        ),
        value=value,
        style={
            "border-width": "0 0 0 2px",
            "border-image": rx.cond(
                index < ToolbarState.step,
                "linear-gradient(180deg, var(--green-9) 0%, var(--green-9) 100%)",
                rx.cond(
                    index == ToolbarState.step,
                    "linear-gradient(180deg, var(--green-9) 0%, var(--slate-3) 100%)",
                    "linear-gradient(180deg, var(--slate-5) 100%, var(--slate-5) 100%)",
                ),
            ),
            "border-image-slice": "1",
            "border-style": "solid",
        },
        class_name="relative !mt-0 [&:not(:last-child)]:pb-4 pl-4 !overflow-visible",
        opacity=rx.cond(index > ToolbarState.step, 0.5, 1),
        pointer_events=rx.cond(index > ToolbarState.step, "none", "auto"),
    )


def accordion() -> rx.Component:
    return rx.box(
        rx.box(
            rx.accordion.root(
                accordion_item(
                    "Selected Code", "code-xml", selected_code_box(), "selected_code", 0
                ),
                accordion_item(
                    "Processing", "ellipsis", process_box(), "processing", 1
                ),
                accordion_item(
                    "Review Changes", "eye", diff_box(), "review_changes", 2
                ),
                variant="ghost",
                collapsible=True,
                default_value="selected_code",
                value=ToolbarState.current_step,
                class_name="pl-4 !rounded-none max-h-full",
            ),
            class_name="overflow-y-auto h-full max-h-full",
        ),
        # Bottom bar
        bottom_bar(),
        class_name="relative flex flex-col justify-between gap-4 border-[--slate-4] bg-[--slate-1] shadow-large p-4 pb-2 border rounded-[1.25rem] w-[300px] md:w-[350px] lg:w-[375px] xl:w-[400px] 2xl:w-[425px] h-[calc(100vh-2rem)] max-h-full resize-x overflow-auto max-w-[80vw] min-w-[300px]",
    )



"""The agent that runs on the user's local client."""

import dataclasses
import difflib
import json
import os

import black
import httpx
from flexai import Agent
from flexai.message import Message
from flexai.tool import ToolCall, ToolResult, parse_tool_item
from pydantic import BaseModel
from reflex.utils import console
from reflex_cli.utils import hosting

import reflex as rx
from reflex_ai.utils import ast_utils, codegen, path_utils
from reflex_ai import paths


# NOTE: using BaseModel here instead of rx.Base due to FastAPI not liking the v1 models
class InternRequest(BaseModel):
    """The request to the AI intern."""

    prompt: str
    selected_code: str
    selected_module: str
    selected_function: str
    parent_component: dict[str, str]


class InternResponse(BaseModel):
    """The response from the AI intern."""

    request_id: str
    messages: list[Message]


class ToolRequestResponse(BaseModel):
    """The response from the tool to the AI intern."""

    request_id: str
    messages: list[ToolResult]


class EditResult(BaseModel):
    """The result of an edit."""

    request_id: str
    diff: str
    accepted: bool


# Tools mapping to human readable names
TOOLS_HR_NAMES = {
    ast_utils.get_module_source.__name__: "Analyzing code",
    codegen.add_python_element.__name__: "Adding new elements",
    codegen.update_python_element.__name__: "Updating elements",
    codegen.delete_python_element.__name__: "Deleting elements",
}

FLEXGEN_BACKEND_URL = os.getenv("FLEXGEN_BACKEND_URL", "http://localhost:8000")


class Diff(rx.Base):
    """A diff between two files (before and after)."""

    # The name of the file
    filename: str

    # The diff of the file
    diff: str


@dataclasses.dataclass()
class LocalAgent:
    """The local agent instance."""

    # The remote backend URL.
    backend_url: str = FLEXGEN_BACKEND_URL

    # The agent instance.
    agent = Agent(
        tools=[
            ast_utils.get_module_source,
            codegen.add_python_element,
            codegen.update_python_element,
            codegen.delete_python_element,
        ],
    )

    @staticmethod
    def authorization_header(token: str) -> dict[str, str]:
        """Construct an authorization header with the specified token as bearer token.

        Args:
            token: The access token to use.

        Returns:
            The authorization header in dict format.
        """
        return {"Authorization": f"Bearer {token}"}

    async def make_request(
        self,
        endpoint: str,
        data: dict,
        timeout: int = 60,
    ) -> dict:
        """Make a request to the backend.

        Args:
            endpoint: The endpoint to send the request to.
            data: The data to send.
            timeout: The timeout for the request.

        Returns:
            The JSON response from the backend.
        """
        token, _ = hosting.get_existing_access_token()
        headers = self.authorization_header(token)

        async with httpx.AsyncClient() as client:
            resp = await client.post(
                f"{self.backend_url}/api/{endpoint}",
                data=data,
                headers=headers,
                timeout=timeout,
            )

        resp.raise_for_status()
        return resp.json()

    async def process(self, request: InternRequest):
        """Process the request from the AI intern.

        Args:
            request: The request from the AI intern.

        Yields:
            The tool use messages from the intern.
        """
        # Create the initial edit request and yield the edit id.
        response = await self.make_request("intern", request.model_dump_json())
        resp_obj = InternResponse(**response)
        yield resp_obj.request_id

        # Process and yield the tool use messages.
        messages = [Message(role=m.role, content=m.content) for m in resp_obj.messages]
        while True:
            tool_results = []
            async for tool_result in self._process_messages(messages, request):
                if isinstance(tool_result, ToolResult):
                    tool_results.append(tool_result)
                else:
                    yield tool_result
            if not tool_results:
                break

            tool_response_request = ToolRequestResponse(
                request_id=resp_obj.request_id,
                messages=tool_results,
            )
            response = await self.make_request(
                "intern/tool_response", tool_response_request.model_dump_json()
            )
            messages = [Message(**m) for m in response]

    async def _process_messages(self, messages: list[Message], request: InternRequest):
        """Process messages from the intern.

        Args:
            messages: The messages from the intern.
            request: The original request.

        Yields:
            ToolCall and ToolResult instances.
        """
        for message in messages:
            if not isinstance(message.content, list):
                continue

            for entry in message.content:
                full_item = parse_tool_item(entry)

                if isinstance(full_item, ToolCall):
                    yield full_item
                    tool_result = await self.agent.invoke_tool(full_item)
                    tool_result = await self._handle_tool_result(
                        tool_result, full_item, request
                    )
                    yield tool_result

                elif isinstance(full_item, ToolResult):
                    yield full_item

    async def _handle_tool_result(
        self, tool_result: ToolResult, tool_call: ToolCall, request
    ):
        """Handle the tool response.

        Args:
            tool_result: The result from the tool.
            tool_call: The tool call item.
            request: The original request.

        Returns:
            The processed tool response.
        """
        if not tool_result.is_error and tool_call.name in [
            codegen.add_python_element.__name__,
            codegen.update_python_element.__name__,
            codegen.delete_python_element.__name__,
        ]:
            # The tool is trying to update the source code.
            suggested_ndiff = tool_result.result
            new_source = "".join(difflib.restore(suggested_ndiff, 2))

            try:
                # Validate the new source
                path_utils.validate_source(
                    new_source, request.parent_component["function"]
                )
            except Exception as e:
                # This code has failed: return the error message
                tool_result = ToolResult(
                    tool_result.tool_use_id,
                    e.stderr,
                    execution_time=0,
                    is_error=True,
                )
            else:
                # This code has passed: write it into the file
                with open(request.parent_component["filename"], "w") as file:
                    content = black.format_str(new_source, mode=black.FileMode())
                    file.write(content)
                tool_result = ToolResult(
                    tool_result.tool_use_id,
                    codegen.to_unified_diff(suggested_ndiff),
                    tool_result.execution_time,
                )

        return tool_result

    async def confirm_change(self, edit_id: str, diffs: list[Diff], accepted: bool):
        """Accept or reject the change.

        Args:
            edit_id: The edit ID.
            diffs: The diffs to accept or reject.
            accepted: Whether the change is accepted or rejected.
        """
        console.info(f"Confirming change for {edit_id} (accepted: {accepted})")
        await self.make_request(
            "intern/edit_result",
            data=EditResult(
                request_id=edit_id,
                diff=json.dumps([d.dict() for d in diffs]),
                accepted=accepted,
            ).model_dump_json(),
        )

        if accepted:
            path_utils.commit_scratch_dir(
                paths.base_paths[0], [d.filename for d in diffs]
            )
        else:
            path_utils.create_scratch_dir(paths.base_paths[0], overwrite=True)



import reflex as rx
from reflex_ai.components import button
from reflex_ai.components import code_block
from reflex_ai.components import icon_button
from reflex_ai.components import hint
from reflex_ai.toolbar import ToolbarState


def dialog(trigger: rx.Component, content: rx.Component) -> rx.Component:
    return rx.dialog.root(
        rx.dialog.trigger(
            trigger,
        ),
        rx.dialog.content(
            content,
            class_name="bg-[--white-1] p-4 rounded-[1.625rem] w-[26rem]",
        ),
    )


def confirm_box() -> rx.Component:
    return rx.box(
        rx.text(
            "Do you accept the changes?",
            class_name="font-base text-[--slate-11] text-balance",
        ),
        rx.box(
            hint(
                text="Accept",
                content=dialog(
                    trigger=icon_button("check", variant="success"),
                    content=rx.box(
                        rx.box(
                            rx.text(
                                "Apply the changes",
                                class_name="font-medium text-[--slate-12] text-xl leading-7 tracking-[-0.01875rem]",
                            ),
                            rx.text(
                                "All your current changes will be applied. Are you sure you want to continue?",
                                class_name="font-base text-[--slate-10]",
                            ),
                            class_name="flex flex-col gap-1 w-full font-instrument-sans",
                        ),
                        rx.box(
                            rx.dialog.close(
                                button(
                                    "Apply",
                                    variant="success",
                                    on_click=lambda: ToolbarState.confirm_change(True),
                                ),
                            ),
                            rx.dialog.close(
                                button(
                                    "Cancel",
                                    variant="secondary",
                                ),
                            ),
                            class_name="flex flex-row justify-between items-center gap-3 w-full",
                        ),
                        class_name="flex flex-col gap-4",
                    ),
                ),
            ),
            hint(
                text="Reject",
                content=dialog(
                    trigger=icon_button("x", variant="destructive"),
                    content=rx.box(
                        rx.box(
                            rx.text(
                                "Start a new session",
                                class_name="font-medium text-[--slate-12] text-xl leading-7 tracking-[-0.01875rem]",
                            ),
                            rx.text(
                                "All your current changes will be discarded. Are you sure you want to continue?",
                                class_name="font-base text-[--slate-10]",
                            ),
                            class_name="flex flex-col gap-1 w-full font-instrument-sans",
                        ),
                        rx.box(
                            rx.dialog.close(
                                button(
                                    "Continue",
                                    variant="destructive",
                                    on_click=lambda: ToolbarState.confirm_change(False),
                                ),
                            ),
                            rx.dialog.close(
                                button("Cancel", variant="secondary"),
                            ),
                            class_name="flex flex-row justify-between items-center gap-3 w-full",
                        ),
                        class_name="flex flex-col gap-4",
                    ),
                ),
            ),
        ),
        # hint(
        #     text="Regenerate",
        #     content=icon_button(
        #         icon="refresh-ccw",
        #         variant="secondary",
        #         on_click=ToolbarState.process({"prompt": ToolbarState.prompt}),
        #     ),
        # ),
        class_name="flex flex-row justify-end items-end gap-3 mt-2 w-full",
    )


def diff_pill(diff) -> rx.Component:
    return rx.box(
        diff.filename.split("/")[-1],
        class_name="border-[--slate-5] hover:bg-[--slate-3] px-3 py-[0.275rem] border rounded-[62.5rem] font-small text-[--slate-11] cursor:pointer transition-bg cursor-pointer",
        background_color=rx.cond(
            ToolbarState.selected_diff.filename == diff.filename,
            "var(--slate-3)",
            "var(--slate-1)",
        ),
        on_click=ToolbarState.set_selected_diff(diff),
    )


def diff_pill_active(text: str) -> rx.Component:
    return rx.box(
        text,
        class_name="border-[--slate-5] bg-[--slate-3] px-3 py-[0.275rem] border rounded-[62.5rem] font-small text-[--slate-11] cursor:pointer transition-bg cursor-pointer",
    )


def diff_pills_stack() -> rx.Component:
    return rx.box(
        rx.foreach(ToolbarState.diff, lambda diff: diff_pill(diff)),
        class_name="flex flex-row flex-wrap gap-2 max-h-[300px] overflow-y-auto shrink-0",
    )


def diff_box() -> rx.Component:
    return rx.box(
        # TODO: Make a summary of the changes
        # rx.el.ul(
        #     rx.el.li("Modified background color to black"),
        #     rx.el.li("Added a new button"),
        #     rx.el.li("Changed the font size"),
        #     class_name="flex flex-col gap-[0.375rem] p-0 pl-[1.25rem] font-base text-[--slate-11] list-disc",
        # ),
        rx.text(
            ToolbarState.changes_comment,
            class_name="font-base text-[--slate-11] m-h-[100px] overflow-y-auto shrink-0",
        ),
        diff_pills_stack(),
        code_block(
            ToolbarState.selected_diff.diff,
        ),
        confirm_box(),
        class_name="flex flex-col gap-3 w-full h-full",
    )



import reflex as rx
from typing import Literal, Callable

LiteralButtonVariant = Literal["primary", "success", "destructive", "secondary"]

default_button_style = {
    "input:placeholder-shown + &": {
        "opacity": "0.65",
        "cursor": "not-allowed",
        "_hover": {
            "background": "linear-gradient(180deg, var(--violet-9) 0%, var(--violet-10) 100%)"
        },
    },
}

default_class_name = "rounded-full w-6 h-6 inline-flex items-center justify-center cursor-pointer relative"
after_class_name = "after:absolute after:inset-[1px] after:border-t after:rounded-[11px] after:border-white after:opacity-[0.22]"


def get_variant_class(variant: str) -> str:
    return (
        f"bg-gradient-to-b from-[--{variant}-9] to-[--{variant}-10] hover:to-[--{variant}-9] text-white"
        + " "
        + after_class_name
    )


variant_styles = {
    "primary": {
        "class_name": get_variant_class("violet"),
    },
    "success": {
        "class_name": get_variant_class("green"),
    },
    "destructive": {
        "class_name": get_variant_class("red"),
    },
    "secondary": {
        "class_name": "bg-[--slate-4] hover:bg-[--slate-6] text-[--slate-10]",
    },
}


def icon_button(
    icon: str,
    variant: LiteralButtonVariant = "primary",
    onclick: Callable = None,
    style: dict = {},
    class_name: str = "",
    *children,
    **props,
) -> rx.Component:
    return rx.el.button(
        rx.icon(icon, size=16, class_name="shrink-0"),
        onclick=onclick,
        style=default_button_style | style,
        class_name=default_class_name
        + " "
        + variant_styles[variant]["class_name"]
        + " "
        + class_name,
        *children,
        **props,
    )
